local clamp = function(min, X, max)
        if X < min then return max end
        if X > max then return max end
        return X
end



local gui = {}
gui.load = function(self)
	self.__checkboxes = {}
	self.__labels = {}
	self.__buttons = {}
	self.__sliders = {}
	self.__textLines = {}
	self.__textBoxes = {}
	self.__drawables = {'checkboxes', 'labels', 'buttons', 'sliders', 'textLines', 'textBoxes'}
	self.__clickables = {'checkboxes', 'buttons', 'textLines', 'textBoxes', 'sliders'}
	self.__updateables = {'sliders', 'buttons'}
	self.__typeables = {'textLines', 'textBoxes'}
		love.keyboard.setKeyRepeat(0.25, 0.025)
end
gui:load()
local white = {255,255,255}
local black = {0,0,0}
local grey = {}
grey.c7 = {224,224,224}
grey.c6 = {192,192,192}
grey.c5 = {160,160,160}
grey.c4 = {128,128,128}
grey.c3 = {96,96,96}
grey.c2 = {64,64,64}
grey.c1 = {32,32,32}

gui.font = {}
setmetatable(gui.font, {__index = function(t, i) if tonumber(i) then t[i] = love.graphics.newFont(i) return t[i] end end})

gui.checkbox = {value = false, width = 12, height = 12, padding = 2, label = '', color = {border = white, check = white, label = white}, size = 12}
gui.checkbox.clicked = function(self, b)
	if b == 'l' then
		self.value = not self.value
	end
end
gui.checkbox.draw = function(self)
	love.graphics.setColor(self.color.border)
	love.graphics.rectangle('line', self.x, self.y, self.width, self.height)
	if self.value then
		love.graphics.setColor(self.color.check)
		love.graphics.rectangle('fill', self.x+self.padding, self.y+self.padding, self.width-self.padding*2, self.height-self.padding*2)
	end
	love.graphics.setColor(self.color.label)
	love.graphics.setFont(gui.font[self.size])
	love.graphics.print(self.label, self.x+self.width+self.padding, self.y+self.padding)
end

gui.delete = function(self, obj)
	table.remove(self['__' .. obj.type], obj.id)
end
	
gui.label = {label = '', color = white, size = 12}
gui.label.draw = function(self)
	love.graphics.setFont(gui.font[self.size])
	love.graphics.print(self.label, self.x, self.y)
end

gui.button = {label = '', padding = 2, color = {border = grey.c6, center = grey.c4, down = grey.c2, up = grey.c4, label = white}, size = 12, clicked = function() end}
gui.button.draw = function(self)
	love.graphics.setColor(self.color.center)
	love.graphics.rectangle('fill', self.x+1, self.y+1, self.width-2, self.height-2)
	love.graphics.setColor(self.color.border)
	love.graphics.rectangle('line', self.x, self.y, self.width, self.height)
	love.graphics.setColor(self.color.label)
	love.graphics.setScissor(self.x+1+self.padding, self.y+1+self.padding, self.width-2-self.padding*2, self.height-2-self.padding*2)
	love.graphics.setFont(gui.font[self.size])
	love.graphics.print(self.label, self.x+1+self.padding, self.y+1+self.padding)
	love.graphics.setScissor()
end
gui.button.update = function(self, b, x, y)
	if x >= self.x and x <= self.x+self.width and y >= self.y and y <= self.y+self.height and b == 'l' then
		self.color.center = self.color.down
	else
		self.color.center = self.color.up
	end
end

gui.slider = {min = 0, max = 100, height = 3, slider = {width = 8, height = 3}, color = {line = grey.c4, slider = grey.c6}}
gui.slider.draw = function(self)
	if not self.vert then
		love.graphics.setLine(self.height, 'rough')
		love.graphics.setColor(self.color.line)
		love.graphics.line(self.x+self.height/2, self.y+self.height/2, self.x+self.width, self.y+self.height/2)
		love.graphics.setColor(self.color.slider)
		love.graphics.rectangle('fill', self.x+(self.value*self.width/self.max)-self.slider.width/2, self.y, self.slider.width, self.slider.height)
	else
		love.graphics.setLine(self.width, 'rough')
		love.graphics.setColor(self.color.line)
		love.graphics.line(self.x+self.width/2, self.y+self.width/2, self.x+self.width/2, self.y+self.height)
		love.graphics.setColor(self.color.slider)
		love.graphics.rectangle('fill', self.x, self.y+(self.value*self.height/self.max)-self.slider.height/2, self.slider.width, self.slider.height)
	end
end
gui.slider.clicked = function(self, b, x, y)
	if x >= self.x and x <= self.x+self.width and y >= self.y and y <= self.y+self.height and b == 'l' then
		self.move = true
	end
end
gui.slider.update = function(self, b, x, y)
	if b ~= 'l' then
		self.move = false
	end
	if not self.vert then
		if self.move and x - self.x >= 0 then
			self.value = clamp(self.min, (x - self.x)*self.max/self.width, self.max)
			self.value = math.floor(self.value + 0.5)
		elseif self.move and x-self.x < 0 then
			self.value = math.max(self.min, (x-self.x)*self.max/self.width)
		end
	else
		if self.move and y - self.y >= 0 then
			self.value = clamp(self.min, (y - self.y)*self.max/self.height, self.max)
			self.value = math.floor(self.value + 0.5)
		elseif self.move and y - self.y < 0 then
			self.value = math.max(self.min, (y - self.y)*self.max/self.width)
		end
	end
end

gui.textLine = {label = '', padding = 2, color = {border = grey.c6, center = grey.c4, label = white}, size = 12, enter = function() end, cursorTime = 0}
gui.textLine.draw = function(self)
	love.graphics.setLine(1, 'smooth')
	love.graphics.setColor(self.color.center)
	love.graphics.rectangle('fill', self.x+1, self.y+1, self.width-2, self.height-2)
	love.graphics.setColor(self.color.border)
	love.graphics.rectangle('line', self.x, self.y, self.width, self.height)
	love.graphics.setColor(self.color.label)
	love.graphics.setScissor(self.x-1+self.padding, self.y+1+self.padding, self.width+2-self.padding*2, self.height-2-self.padding*2)
	love.graphics.setFont(gui.font[self.size])
	local printX = self.x+1+self.padding
	local labelWidth = love.graphics.getFont():getWidth(self.label)
	local cursorX = love.graphics.getFont():getWidth(self.label:sub(1, self.cursor)) + self.padding + printX
	if cursorX < self.x + 1 + self.padding then
		printX = printX + (self.x + 1 + self.padding) - cursorX
		cursorX = self.x + 1 + self.padding
	end
	if cursorX > self.x - 2 - self.padding*2 + self.width then
		printX = printX - ((self.x + 2 - self.padding*2) - self.width) - cursorX
		cursorX = self.x + 2 - self.padding*2 + self.width
	end
	love.graphics.print(self.label, printX, self.y+1+self.padding)
	if self.cursorTime < 0.5 and self.focus then
		love.graphics.setLine(1, 'rough')
		love.graphics.line(cursorX, 1, cursorX, 1080)
	end
	love.graphics.setScissor()
end
gui.textLine.clicked = function(self, b)
	if b == 'l' then
		self.focus = true
	end
end
gui.textLine.typed = function(self, key, uni)
	if uni > 27 and uni < 127 then
		local p1, p2 = self.label:sub(1, self.cursor), self.label:sub(self.cursor+1, -1)
		self.label = p1 .. string.char(uni) .. p2
		self.cursor = self.cursor + 1
	end
	if key == 'left' then
		self.cursor = math.max(self.cursor - 1, 0)
	end
	if key == 'right' then
		self.cursor = math.min(self.cursor + 1, #self.label)
	end
	if uni == 8 and self.cursor ~= 0 then
		self.label = self.label:sub(1, self.cursor-1) .. self.label:sub(self.cursor+1, -1)
		self.cursor = self.cursor-1
	end
	if uni == 127 and self.cursor ~= #self.label then
		self.label = self.label:sub(1, self.cursor) .. self.label:sub(self.cursor+2, -1)
	end
	if key == 'home' then
		self.cursor = 0
	end
	if key == 'end' then
		self.cursor = #self.label
	end
	if key == 'return' then
		self:enter()
	end
end

gui.textBox = {text = '', padding = 2, color = {border = grey.c6, center = grey.c4, label = white}, size = 12, cursorTime = 0}
gui.textBox.draw = function(self)
	love.graphics.setLine(1, 'smooth')
	love.graphics.setColor(self.color.center)
	love.graphics.rectangle('fill', self.x+1, self.y+1, self.width-2, self.height-2)
	love.graphics.setColor(self.color.border)
	love.graphics.rectangle('line', self.x, self.y, self.width, self.height)
	love.graphics.setScissor(self.x-1+self.padding, self.y+1+self.padding, self.width+2-self.padding*2, self.height-2-self.padding*2)
	love.graphics.setFont(gui.font[self.size])
	if self.cursorTime < 0.5 and self.focus then
		love.graphics.setColor(self.color.label[1], self.color.label[2], self.color.label[3], (self.color.label[4] or 255)/1.75)
		love.graphics.printf(self.text:sub(1, self.cursor.pos) .. '|', self.x-self.topLine*love.graphics.getFont():getHeight()+1+self.padding, self.y+1+self.padding, self.width-1-self.padding, 'left')
	else
		love.graphics.setColor(self.color.label)
	end
	love.graphics.printf(self.text, self.x-self.topLine*love.graphics.getFont():getHeight()+1+self.padding, self.y+1+self.padding, self.width-1-self.padding, 'left')
	love.graphics.setScissor()
end
gui.textBox.clicked = function(self, b)
	if b == 'l' then
		self.focus = true
	end
end
gui.textBox.typed = function(self, key, uni)
	if uni > 27 and uni < 127 then
		local p1, p2 = self.text:sub(1, self.cursor.pos), self.text:sub(self.cursor.pos+1, -1)
		self.text = p1 .. string.char(uni) .. p2
		self.cursor.pos = self.cursor.pos + 1
	end
	if key == 'left' then
		self.cursor.pos = math.max(self.cursor.pos - 1, 0)
	end
	if key == 'right' then
		self.cursor.pos = math.min(self.cursor.pos + 1, #self.text)
	end
	if uni == 8 and self.cursor ~= 0 then
		self.text = self.text:sub(1, self.cursor.pos-1) .. self.text:sub(self.cursor.pos+1, -1)
		self.cursor.pos = self.cursor.pos-1
	end
	if uni == 127 and self.cursor ~= #self.text then
		self.text = self.text:sub(1, self.cursor.pos) .. self.text:sub(self.text.pos+2, -1)
	end
	if key == 'home' then
		self.cursor.pos = 0
	end
	if key == 'end' then
		self.cursor.pos = #self.text
	end
	if key == 'return' then
		self.text = self.text .. "\n"
		self.cursor.pos = self.cursor.pos + 1
	end
end

gui.newCheckbox = function(self, x, y, v, l, w, h, p, c, s)
	local t = {}
	t.x = x
	t.y = y
	t.value = v
	t.label = l
	t.width = w
	t.height = h
	t.padding = p
	t.color = c
	t.size = s
	t.id = #self.__checkboxes + 1
	t.type = 'checkboxes'
	setmetatable(t, {__index = function(t, i) return self.checkbox[i]end})
	table.insert(self.__checkboxes, t)
	return self.__checkboxes[t.id]
end

gui.newLabel = function(self, x, y, l, c, s)
	local t = {}
	t.x = x
	t.y = y
	t.label = l
	t.color = c
	t.size = s
	t.id = #self.__labels + 1
	t.type = 'labels'
	setmetatable(t, {__index = function(t,i) return self.label[i] end})
	table.insert(self.__labels, t)
	return self.__labels[t.id]
end

gui.newButton = function(self, x, y, f, l, w, h, p, c, s)
	local t = {}
	t.x = x
	t.y = y
	t.clicked = f
	t.label = l
	t.padding = p or self.button.padding
	t.color = c or {border = {}, center = {}, down = {}, up = {}, label = {}}
	t.size = s
	t.width = w or self.font[s or self.button.size]:getWidth(l or '')+t.padding*2+2
	t.height = h or self.font[s or self.button.size]:getHeight()+t.padding*2+2
	t.id = #self.__buttons + 1
	t.type = 'buttons'
	if not c then
		for i, v in pairs(t.color) do
			v[1] = gui.button.color[i][1]
			v[2] = gui.button.color[i][2]
			v[3] = gui.button.color[i][3]
		end
	end
	setmetatable(t, {__index = function(t,i) return self.button[i] end})
	table.insert(self.__buttons, t)
	return self.__buttons[t.id]
end

gui.newSlider = function(self, x, y, min, max, value, vert, w, h, slider, color)
	local t = {}
	t.x = x
	t.y = y
	t.min = min
	t.max = max
	t.value = value or (min+max)/2
	t.vert = vert
	t.height = h
	t.slider = slider
	t.color = color
	t.width = w or max
	t.id = #self.__sliders + 1
	t.type = 'sliders'
	setmetatable(t, {__index = function(t,i) return self.slider[i] end})
	table.insert(self.__sliders, t)
	return self.__sliders[t.id]
end

gui.newTextLine = function(self, x, y, enter, label, width, size, padding, color)
	local t = {}
	t.x = x
	t.y = y
	t.enter = enter
	t.label = label
	t.width = width or gui.font[size or self.textLine.size]:getWidth(t.label or self.textLine.label)
	t.height = gui.font[size or self.textLine.size]:getHeight() + (padding or self.textLine.padding)*2 + 2
	t.size = size
	t.padding = padding
	t.color = color or {border = {}, center = {}, label = {}}
	t.cursor = #t.label
	t.id = #self.__textLines + 1
	t.type = 'textLines'
	if not color then
		for i, v in pairs(t.color) do
			v[1] = gui.textLine.color[i][1]
			v[2] = gui.textLine.color[i][2]
			v[3] = gui.textLine.color[i][3]
		end
	end
	setmetatable(t, {__index = function(t,i) return self.textLine[i] end})
	table.insert(self.__textLines, t)
	return self.__textLines[t.id]
end

gui.newTextBox = function(self, x, y, text, width, height, padding, color, size)
	local t = {}
	t.x = x
	t.y = y
	t.text = text
	t.width = width
	t.height = height
	t.padding = padding
	t.color = color or {border = {}, center = {}, label = {}}
	t.cursor = {}
	t.cursor.pos = #t.text
	t.cursor.x = #t.text % t.width
	t.cursor.y = math.floor(#t.text / t.width)
	t.topLine = 0
	t.id = #self.__textBoxes + 1
	t.type = 'textBoxes'
	if not color then
		for i, v in pairs(t.color) do
			v[1] = self.textBox.color[i][1]
			v[2] = self.textBox.color[i][2]
			v[3] = self.textBox.color[i][3]
		end
	end
	setmetatable(t, {__index = function(t,i) return self.textBox[i] end})
	table.insert(self.__textLines, t)
	return self.__textLines[t.id]
end

gui.update = function(self, dt)
	self.textLine.cursorTime = self.textLine.cursorTime + dt
	self.textBox.cursorTime = self.textBox.cursorTime + dt
	if self.textLine.cursorTime >= 1 then
		self.textLine.cursorTime = self.textLine.cursorTime - 1
	end
	if self.textBox.cursorTime >= 1 then
		self.textBox.cursorTime = self.textBox.cursorTime - 1
	end
	local x, y = love.mouse.getPosition()
	local b = ''
	if love.mouse.isDown 'l' then b = 'l'
	elseif love.mouse.isDown 'r' then b = 'r'
	elseif love.mouse.isDown 'm' then b = 'm'
	elseif love.mouse.isDown 'wu' then b = 'wu'
	elseif love.mouse.isDown 'wd' then b = 'wd'
	elseif love.mouse.isDown 'x1' then b = 'x1'
	elseif love.mouse.isDown 'x2' then b = 'x2'
	else b = nil
	end
	for i, v in pairs(self.__updateables) do
		for i, v in ipairs(self['__' .. v]) do
			v:update(b, x, y)
		end
	end
end

gui.mousepressed = function(self, x, y, b)
	for i, v in pairs(self.__clickables) do
		if v == 'textLines' then
			reset = function(v) v.focus = false end
		else
			reset = function() end
		end
		for i, v in ipairs(self['__' .. v]) do
			reset(v)
			if x >= v.x and x <= v.x+v.width and y >= v.y and y <= v.y+v.height then
				v:clicked(b, x, y)
			end
		end
	end
end

gui.keypressed = function(self, key, uni)
	for i, v in ipairs(self.__typeables) do
		for i, v in ipairs(self['__' .. v]) do
			if v.focus then
				self.textLine.cursorTime = 0
				v:typed(key, uni)
			end
		end
	end
end

gui.draw = function(self)
	local settings = {}
	love.graphics.setLine(1, 'smooth')
	for i, v in pairs(self.__drawables) do
		for i, v in ipairs(self['__' .. v]) do
			v:draw()
		end
	end
end

gui.erase = function(self)
	self:load()
end

return gui